use crate::{definition::Scope, server::LSPServer, sources::{LSPSupport, ParseIR, Source}, utils::{get_definition_token, get_language_id_by_uri}};
use log::info;
use sv_parser::{RefNode, SyntaxTree};
use vhdl::vhdl_document_highlight;
use crate::definition::extract_defs::get_ident;
use tower_lsp::lsp_types::*;

pub mod vhdl;

impl LSPServer {
    pub fn document_highlight(
        &self,
        params: DocumentHighlightParams,
    ) -> Option<Vec<DocumentHighlight>> {
        let uri = &params.text_document_position_params.text_document.uri;
        let pos = params.text_document_position_params.position;
        let file_id = self.srcs.get_id(&uri).to_owned();
        self.srcs.wait_parse_ready(file_id, false);
        let file = self.srcs.get_file(file_id)?;
        let file = file.read().ok()?;
        let line_text = file.text.line(pos.line as usize);
        let token = get_definition_token(&line_text, pos);

        let language_id = get_language_id_by_uri(&uri);
        
        match language_id.as_str() {
            "vhdl" => vhdl_document_highlight(self, &params.text_document_position_params),
            "verilog" | "systemverilog" => sv_document_highlight(
                self,
                &token,
                &file,
                pos,
                &uri
            ),
            _ => None
        }
    }
}

fn sv_document_highlight(
    server: &LSPServer,
    token: &str,
    file: &Source,
    pos: Position,
    uri: &Url
) -> Option<Vec<DocumentHighlight>> {
    let scope_tree = server.srcs.scope_tree.read().ok()?;

    // use the byte_idx of the definition if possible, otherwise use the cursor
    let byte_idx =
        match scope_tree
            .as_ref()?
            .get_definition(token, file.text.pos_to_byte(&pos), uri)
        {
            Some(def) => def.byte_idx,
            None => file.text.pos_to_byte(&pos),
        };
    let syntax_tree = file.parse_ir.as_ref()?;
    match syntax_tree {
        ParseIR::SyntaxTree(syntax_tree) => {
            let references = all_identifiers(&syntax_tree, &token);
            Some(
                scope_tree
                    .as_ref()?
                    .document_highlights(&uri, &file.text, references, byte_idx),
            )
        },
        _ => {
            info!("error happen in [sv_document_highlight]");
            None
        }
    }
}



/// return all identifiers in a syntax tree matching a given token
fn all_identifiers(syntax_tree: &SyntaxTree, token: &str) -> Vec<(String, usize)> {
    let mut idents: Vec<(String, usize)> = Vec::new();
    for node in syntax_tree {
        if let RefNode::Identifier(_) = node {
            
            let (ident, byte_idx) = get_ident(syntax_tree, node);
            if ident == token {
                idents.push((ident, byte_idx));
            }
        }
    }
    idents
}